/*
 * Copyright (c) 2014 - Zephyr Software LLC
 *
 * This file may be used and modified for non-commercial purposes as long as
 * all copyright, permission, and nonwarranty notices are preserved.
 * Redistribution is prohibited without prior written consent from Zephyr
 * Software.
 *
 * Please contact the authors for restrictions applying to commercial use.
 *
 * THIS SOURCE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author: Zephyr Software
 * e-mail: jwd@zephyr-software.com
 * URL   : http://www.zephyr-software.com/
 *
 */

#include <zipr_all.h>
#include <unistd.h>
#include <stdlib.h>
#include <getopt.h>
#include <algorithm>    // std::random_shuffle
#include <vector>       // std::vector
#include <ctime>        // std::time
#include <cstdlib>      // std::rand, std::srand


using namespace zipr;
using namespace std;

ZiprOptionsNamespace_t *ZiprMemorySpace_t::RegisterOptions(ZiprOptionsNamespace_t *global) {
	global->AddOption(&m_verbose);
	return NULL;
}

void ZiprMemorySpace_t::SplitFreeRange(Range_t split_from)
{
	RangeAddress_t counter, end;
	for (counter = split_from.GetStart(), end = split_from.GetEnd();
	     counter!=end;
			 counter++)
	{
		SplitFreeRange(counter);
	}
}

void ZiprMemorySpace_t::SplitFreeRange(RangeAddress_t addr)
{
	RangeSet_t::iterator it=FindFreeRange(addr);
	assert(IsValidRange(it));

	Range_t r=*it;
	if(r.GetStart()==r.GetEnd())
	{
		assert(addr==r.GetEnd());
		free_ranges.erase(it);
	}
	else if(addr==r.GetStart())
	{
		free_ranges.erase(it);
		free_ranges.insert(Range_t(r.GetStart()+1, r.GetEnd()));
	}
	else if(addr==r.GetEnd())
	{
		free_ranges.erase(it);
		free_ranges.insert(Range_t(r.GetStart(), r.GetEnd()-1));
	}
	else // split range 
	{
		free_ranges.erase(it);
		free_ranges.insert(Range_t(r.GetStart(), addr-1));
		free_ranges.insert(Range_t(addr+1, r.GetEnd()));
	}
}
void ZiprMemorySpace_t::MergeFreeRange(Range_t range)
{
	RangeAddress_t counter, end;
	for (counter = range.GetStart(), end = range.GetEnd();
	     counter!=end;
			 counter++)
	{
		MergeFreeRange(counter);
	}
}
void ZiprMemorySpace_t::MergeFreeRange(RangeAddress_t addr)
{
	/*
	 * Make a new range of one byte.
	 * 
	 * Then, look to see whether or not it
	 * can be merged with another range.
	 *
	 * If not, add it as a one byte range.
	 */

	Range_t nr(addr, addr);
	Range_t nrp1(addr+1, addr+1);
	Range_t nrm1(addr-1, addr-1);
	RangeSet_t::iterator itp1;
	RangeSet_t::iterator itm1;

	itp1=free_ranges.find(nrp1);
	itm1=free_ranges.find(nrm1);
	Range_t r;
	if(itp1!=free_ranges.end())
	{
		r=*itp1;
		assert((addr+1) == r.GetStart());
		/*
		 * Make the beginning of this range
		 * one byte smaller!
		 */
		Range_t nnr(addr, r.GetEnd());
		if (m_verbose)
		{
			printf("Expanded range: ");
			printf("from: (%p - %p) ", (void*)r.GetStart(), (void*)r.GetEnd());
			printf("to: (%p - %p) \n", (void*)nnr.GetStart(), (void*)nnr.GetEnd());
		}
		nr = nnr;
		/*
		 * There is the possibility that we just expanded down and we
		 * now abut the end of a smaller free one.
		 */
		if(itm1!=free_ranges.end())
		{
			Range_t r2=*itm1;
			if (m_verbose)
			{
				printf("Expanded range: ");
				printf("from: (%p - %p) ", (void*)nr.GetStart(), (void*)nr.GetEnd());
				printf("to: (%p - %p) \n", (void*)r2.GetStart(), (void*)nnr.GetEnd());
			}
			nr.SetStart(r2.GetStart());	
			free_ranges.erase(r2);
			
		}
		/*
		 * Handling the analagous secondary merge case
		 * does not need to be handled here -- that 
		 * would have been taken care of in the 1st 
		 * conditional (just using opposite terms).
		 */
		free_ranges.erase(r);
	} 
	else if(itm1!=free_ranges.end())	 // not addr+1 is still busy, so we don't merge against it.
	{
		r=*itm1;
	
		assert((addr-1) == r.GetEnd()) ;
		/*
		 * Make the end of this range one byte
		 * bigger
		 */
		Range_t nnr(r.GetStart(), addr);
		if (m_verbose)
		{
			printf("Expanded range: ");
			printf("from: (%p - %p) ", (void*)r.GetStart(), (void*)r.GetEnd());
			printf("to: (%p - %p) \n", (void*)nnr.GetStart(), (void*)nnr.GetEnd());
		}
		nr = nnr;
		free_ranges.erase(itm1);
	}
	else
	{
		// else both p1 and m1 are busy still, so we just insert nr
	}

	// insert NR and be done.
	free_ranges.insert(nr);
	return;

}

void ZiprMemorySpace_t::PrintMemorySpace(std::ostream &out)
{
	for( RangeSet_t::iterator it=free_ranges.begin();
		it!=free_ranges.end();
		++it)
	{
		Range_t r = *it;
		out <<"0x"<<std::hex<<r.GetStart()<<" - 0x"<<std::hex<<r.GetEnd()<<endl;
	}
}

RangeSet_t::iterator ZiprMemorySpace_t::FindFreeRange(RangeAddress_t addr)
{
	RangeSet_t::iterator freer = free_ranges.find(Range_t(addr, addr)); 
	return freer;
}

bool ZiprMemorySpace_t::IsValidRange(RangeSet_t::iterator it)
{
	return it!=free_ranges.end();
}

std::pair<RangeSet_t::const_iterator,RangeSet_t::const_iterator>
	ZiprMemorySpace_t::GetNearbyFreeRanges(const RangeAddress_t hint,size_t count)
{
	Range_t search(hint, hint+1);
	RangeSet_t::const_iterator result = free_ranges.lower_bound(search);
	/*
	 * TODO: Not quite sure what to make of this.
	 */
	return std::pair<RangeSet_t::const_iterator,RangeSet_t::const_iterator>(
		result, free_ranges.end());
}

Range_t ZiprMemorySpace_t::GetLargeRange(void)
{
	for( RangeSet_t::iterator it=free_ranges.begin();
		it!=free_ranges.end();
		++it)
	{
		Range_t r=*it;
		if(r.GetEnd()==(RangeAddress_t)-1)
			return r;
	}
	return Range_t(0,0);
}

bool ZiprMemorySpace_t::SortRangeBySize(const Range_t &a, const Range_t &b)
{
	return (a.GetEnd() - a.GetStart()) <= (b.GetEnd() - b.GetStart());
}

std::list<Range_t> ZiprMemorySpace_t::GetFreeRanges(size_t size)
{
	list<Range_t> result;
	for( RangeSet_t::iterator it=free_ranges.begin();
		it!=free_ranges.end();
		++it)
	{
		Range_t r=*it;
		if(r.GetEnd() - r.GetStart() >= (unsigned) size)
			result.push_back(r);
	}
	result.sort(SortRangeBySize);
	return result;
}

Range_t ZiprMemorySpace_t::GetInfiniteFreeRange()
{
	vector<Range_t> v;
	Range_t big_range;
	for( RangeSet_t::iterator it=free_ranges.begin();
		it!=free_ranges.end();
		++it)
	{
		Range_t r=*it;
		if(r.GetEnd()==(RangeAddress_t)-1)
			return r;
	}
	assert(false);
	return Range_t(0,0);
}

Range_t ZiprMemorySpace_t::GetFreeRange(int size)
{
	vector<Range_t> v;
	Range_t big_range;
	for( RangeSet_t::iterator it=free_ranges.begin();
		it!=free_ranges.end();
		++it)
	{
		Range_t r=*it;
		if(r.GetEnd()==(RangeAddress_t)-1)
			big_range=r;
		else if(r.GetEnd() - r.GetStart() >= (unsigned) size)
			v.push_back(r);

		// that's enough randomization
		if(v.size() > 100)
			break;
	}
	if(v.size()==0)
		return big_range;	

	// choose random value to return.
	int index=std::rand() % v.size();
	return v[index];
}

// queries about free areas.
bool ZiprMemorySpace_t::AreBytesFree(RangeAddress_t addr, int num_bytes)
{
	for(int i=0;i<num_bytes;i++)
		if(!IsByteFree(addr+i))
			return false;
	return true;
}

bool ZiprMemorySpace_t::IsByteFree(RangeAddress_t addr)
{
	if (IsValidRange(FindFreeRange(addr)))
		return true;
	return false;
}

void ZiprMemorySpace_t::AddFreeRange(Range_t newRange, bool original)
{
	if (original)
	{
		original_free_ranges.insert(Range_t(newRange.GetStart(),newRange.GetEnd()));
	}
	AddFreeRange(newRange);
}

void ZiprMemorySpace_t::AddFreeRange(Range_t newRange)
{
	free_ranges.insert(Range_t(newRange.GetStart(), newRange.GetEnd()));
}
void ZiprMemorySpace_t::RemoveFreeRange(Range_t oldRange)
{
	free_ranges.erase(Range_t(oldRange.GetStart(), oldRange.GetEnd()));
}

int ZiprMemorySpace_t::GetRangeCount()
{
	return free_ranges.size();
}
